# -*- coding: utf-8 -*-
"""
Created on Sun Oct  2 21:28:02 2022

@author: lidon
"""

from vector import*
from linked_list import*


class queue(linked_list):
    # insert e to the last
    def enqueue(self,e):
        self.insert_as_last(e)
        
    # remove the first one
    def dequeue(self):
        return self.remove(self.first())
    def front(self):
        return self.first().data


# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:31:05 2022

@author: lidon
"""
# implementation of binary searching tree
from bin_tree import*
from bin_node import*
from vector import*
from linked_list import*


class entry:
    def __init__(self,key=None,value=None):
        self.key=key
        self.value=value
        
    def __lt__(self,other):
        return self.key<other.key
    def __gt__(self,other):
        return self.key>other.key
    def __le__(self,other):
        return self.key<=other.key
    def __ge__(self,other):
        return self.key>=other.key
    def __eq__(self,other):
        return self.key==other.key
    def __ne__(self,other):
        return self.key!=other.key
    
    def __repr__(self):
        return str([self.key,self.value])


class bst(bin_tree):
    # search in the bst starting from v for key e
    # hot is auxillary var
    def search_in(self,v,e,hot):
        if (not v) or (v.data==e):
            return v,hot
        hot=v
        if e<v.data:
            return self.search_in(v.lc,e,hot)
        else:
            return self.search_in(v.rc,e,hot)
    
    # search for e in bst
    def search(self,e):
        v,hot=self.search_in(self.root,e,hot=None)
        return v
    
    # insert e
    def insert(self,e):
        v,hot=self.search_in(self.root,e,hot=None)
        # the node already exists
        if v:
            return v
        x=bin_node(data=e,par=hot)
        if e<hot.data:
            hot.lc=x
        else:
            hot.rc=x
        self.size=self.size+1
        self.update_height_above(x)
        # return the inserted node
        return x
    
    # remove node v
    def remove_at(self,v,hot):
        w=v
        succ=None
        if not has_lc(v):
            v=v.rc
            succ=v
        elif not has_rc(v):
            v=v.lc
            succ=v
        else:
            w=w.succ()
            v.data,w.data=swap(v.data,w.data)
            u=w.par
            if u==v:
                succ=w.rc
                u.rc=succ
            else:
                succ=w.rc
                u.lc=succ
        hot=w.par
        if succ:
            succ.par=hot
            if hot:
                if succ.data<hot.data:
                    hot.lc=succ
                else:
                    hot.rc=succ
        # return the successor
        return succ
    
    # remove data e
    def remove(self,e):
        v,hot=self.search_in(self.root,e,hot=None)
        self.remove_at(v,hot)
        self.size=self.size-1
        self.update_height_above(hot)
        return True
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Sep 25 11:29:16 2022

@author: lidon
"""

# Abstract Data Type of Vector

class vector:
    # arr should be a python list 
    def __init__(self,arr):
        self.elements=arr
        self.length=len(arr)
    
    # reload the [] operator
    def __getitem__(self,key):
        return self.elements[key]
    
    # reload setitem function
    def __setitem__(self,key,value):
        self.elements[key]=value
    
    # reload the representation to print
    def __repr__(self):
        return self.elements.__repr__()
    
    # reload len(x)
    def __len__(self):
        return self.elements.__len__()
    
    def __add__(self,other):
        new_elements=self.elements+other.elements
        return vector(new_elements)
    
    # return length of the vector
    def size(self):
        pass
    
    # return if the vector is empty
    def empty(self):
        return not self.length>0
    
    # return if the vector is in order
    # return total number of disordered pairs
    def disordered(self):
        pass
    
    # find element e from [low,high] vector, return its index
    def find(self, e,low=None,high=None):
        pass
    
    # search e using binary search
    def bin_search(self,e,low=None,high=None):
        pass
    
    # search e from the vector, assume vector is ordered (find doesn't assume so)
    def search(self,e,func="bin_search",low=None,high=None):
        if func=="bin_search":
            return self.bin_search(e,low,high)
    
    # exchange elements
    def bubble(self,low=None,high=None):
        pass
    
    # bubble sort
    def bubble_sort(self,low=None,high=None):
        pass
    
    # return the maximum number
    def maximum(self,low=None,high=None):
        pass
    
    # selection sort
    def selection_sort(self,low=None,high=None):
        pass
    
    # merge, used for merge sort
    def merge(self,low,mid,high):
        pass
    
    # merge sort
    def merge_sort(self,low=None,high=None):
        pass
    
    # line sweep
    def partition(self,low=None,high=None):
        pass
    
    # quick sort
    def quick_sort(self,low=None,high=None):
        pass
    
    # heap sort
    def heap_sort(self,low=None,high=None):
        pass
    
    # additionally define copy, to match the syntax of python
    # deep copy, instead of setting the pointer
    def copy(self):
        pass
    
    #remove element in index r
    def remove(self,r):
        pass
    
    # remove elements in [low,high)
    def remove_index(self,low,high):
        pass
    
    # insert element e into position r
    def insert(self,r,e):
        pass
    
    # insert element e at the tail of the vector
    def append(self,e):
        self.insert(self.length,e)
    
    # sort elements in [low,high)
    # if don't specify low and high, sort the whole vector
    def sort(self,func="quick",low=None,high=None):
        if func=="bubble":
            return self.bubble_sort(low,high)
        elif func=="selection":
            return self.selection_sort(low,high)
        elif func=="merge":
            return self.merge_sort(low,high)
        elif func=="heap":
            return self.heap_sort(low,high)
        elif func=="quick":
            return self.quick_sort(low,high)
    
    # unsort elements in [low,high) by permuting
    # if don't specify low and high, unsort the whole vector
    def unsort(self,low=None,high=None):
        pass
    
    # remove repeating elements
    # assume vector unsorted
    def deduplicate(self):
        pass
    
    # remove repeating elements
    # assume vector sorted
    def uniquify(self):
        pass
    
    # traverse the whole vector
    # apply func on every element in the vector
    def traverse(self,func):
        pass
    
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Oct  2 22:22:25 2022

@author: lidon
"""
from queues import*
from bin_node import*



class bin_tree:
    # size: tree size
    # root: tree root
    def __init__(self,size=0,root=None):
        self.size=size
        self.root=root
        
    # update the height of node x
    # return the updated height
    def update_height(self,x):
        x.height=1+max(stature(x.lc),stature(x.rc))
        return x.height
    
    # update height of x and its ancestors
    def update_height_above(self,x):
        while x:
            self.update_height(x)
            x=x.par
    
    # return size of the tree
    def size(self):
        return self.size
    
    # empty or not
    def empty(self):
        return not self.root
    
    # return root
    def root(self):
        return self.root
    
    # insert as root with data e
    # return root
    # WARNING: this will remove the full tree if there exists any!
    def insert_as_root(self,e):
        self.size=1
        self.root=bin_node(data=e)
        return self.root
    
    # insert e as the lc of node x
    def insert_as_lc(self,x,e):
        self.size=self.size+1
        x.insert_as_lc(e)
        self.update_height_above(x)
        return x.lc
    
    def insert_as_rc(self,x,e):
        self.size=self.size+1
        x.insert_as_rc(e)
        self.update_height_above(x)
        return x.rc
    
    # attach tree as left subtree of x
    # return position of x
    # WARNING: the original left subtree of x will be removed (if any)!
    def attach_as_lc(self,x,tree):
        x.lc=tree.root
        tree.root.par=x
        self.size=self.size+tree.size
        self.update_height_above(x)
        tree.root=None
        tree.size=0
        return x
    
    def attach_as_rc(self,x,tree):
        x.rc=tree.root
        tree.root.par=x
        self.size=self.size+tree.size
        self.update_height_above(x)
        tree.root=None
        tree.size=0
        return x
    
    # remove subtree with x as its root
    def remove(self,x):
        # if x is the root of the tree
        if isroot(x):
            self.size=0
            self.root=None
        else:
            if is_lc(x):
                x.par.lc=None
            else:
                x.par.rc=None
            self.update_height_above(x.par)
            x.par=None
                    
            
    
    # remove subtree with x as its root and transform it into an independent subtree
    def secede(self,x):
        # if x is the root, just return the tree itself
        if isroot(x):
            return self
        else:
            # cutoff pointer from the parents
            if is_lc(x):
                x.par.lc=None
            else:
                x.par.rc=None
            # renew heights
            self.update_height_above(x.par)
            # construct a new tree starting from x
            new_tree=bin_tree(size=x.size(),root=x)
            x.par=None
            
            # renew size of the old tree
            self.size=self.size-new_tree.size
            return new_tree
    
    # traverse by level
    # x a node in tree
    def trav_level(self,x,func):
        Q=queue([])
        Q.enqueue(x)
        while len(Q)!=0:
            node_now=Q.dequeue()
            func(node_now.data)
            if has_lc(node_now):
                Q.enqueue(node_now.lc)
            if has_rc(node_now):
                Q.enqueue(node_now.rc)

    # preorder traverse
    # x is a node in the tree
    def trav_pre(self,x,func):
        if not x:
            return
        func(x.data)
        self.trav_pre(x.lc,func)
        self.trav_pre(x.rc,func)
    
    # in order traverse
    # x is a node in the tree
    def trav_in(self,x,func):
        if not x:
            return
        self.trav_in(x.lc,func)
        func(x.data)
        self.trav_in(x.rc,func)
    
    # traverse in post order
    # x a node in the tree
    def trav_post(self,x,func):
        if not x:
            return
        self.trav_post(x.lc,func)
        self.trav_post(x.rc,func)
        func(x.data)
    
    
    
    # reload <=
    def __le__(self,other):
        return self.root and  other.root and self.root<=other.root
    
    # reload ==
    def __eq__(self,other):
        return self.root and other.root and self.root==other.root
    




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 26 22:33:02 2022

@author: lidon
"""

class list_node:
    def __init__(self,data=None,pred=None,succ=None):
        self.data=data
        self.pred=pred
        self.succ=succ
    def insert_as_pred(self,e):
        pass
    def insert_as_succ(self,e): 
        pass
    
class linked_list:
    #initialize the list with an array elements
    def __init__(self,elements=None):
        if elements==None:
            self.header=list_node()
            self.trailer=list_node()
            self.header.succ=self.trailer
            self.header.pred=None
            self.trailer.pred=self.header
            self.trailer.succ=None
            self.size=0
        else:
            # avoid using insert as last
            self.header=list_node()
            self.trailer=list_node()
            self.header.pred=None
            self.header.succ=self.trailer
            self.trailer.succ=None
            self.trailer.pred=self.header
            self.size=0
            for i in range(0,len(elements)):
                self.size=self.size+1
                self.trailer.insert_as_pred(elements[i])
            
        
    def __getitem__(self,key):
        p=self.first()
        while key>0:
            p=p.succ
            key=key-1
        return p.data
    
    def __setitem__(self,key,value):
        p=self.first()
        while key>0:
            p=p.succ
            key=key-1
        p.data=value
    
    def __len__(self):
        return self.size
    
    def __repr__(self):
        string=[]
        p=self.first()
        while p.succ!=None:
            string=[string]+[p.data]
            p=p.succ
        string=[string]
        #print(string)
        return str(string)
    
    # copy a linked list
    # from node to its n-1 th pred
    def copy(self,node=None,n=None):
        return self.copy_nodes(node=self.first(),n=self.size)
    
    # remove a node
    # return the removed value
    def remove(self,node):
        pass
        
    # remove all nodes
    # return number of nodes removed
    def clear(self):
        pass
    
    # copy n nodes starting from node to its n-1 succ
    # return a list consisting of these nodes
    def copy_nodes(self,node,n):
        pass
    
    # return size
    def size(self):
        pass
    
    # if it is empty (True if empty)
    def empty(self):
        return self.size<=0
    
    # return first node
    def first(self):
        return self.header.succ
    
    # return last node
    def last(self):
        return self.trailer.pred
    
    # return if node is valid (can be exposed to user)
    def valid(self,node):
        return node and (node!=self.header) and (node!=self.trailer)
    
    # return if ordered
    def disordered(self):
        pass
    
    # find data from linked list
    # from node to its n th pred
    # assume disordered
    def find(self,e,n=None,node=None):
        pass
    
    # search e and return the node <=e with largest rank
    # search e from n pred starting from node (from back to front)
    # assume ordered
    # if n and node are none, search from the whole list
    def search(self,e,n=None,node=None):
        pass
    
    # find the largest among node and its n-1 successors
    # if node and n None, search from whole list
    def select_max(self,node=None,n=None):
        if n==None and node==None:
            n=self.size
            node=self.header.succ
        p=node
        maximum=node
        for r in range(0,n):
            if maximum.data<p.data:
                maximum=p
            p=p.succ
        return maximum
    
    def insert_as_first(self,e):
        pass
    
    def insert_as_last(self,e):
        pass
    
    # insert e as the successor of node
    def insert_succ(self,node,e):
        pass
    
    # insert e as the pred of node
    def insert_pred(self,node,e):
        pass
    
    # remove a node
    def remove(self,node):
        pass
    
    # selection sort
    # sort n values starting from node
    def selection_sort(self,node=None,n=None):
        pass
    
    
    
    # insertion sort
    # sort n values starting from node
    def insertion_sort(self,node=None,n=None):
        pass
    
    # merge function
    # merge n nodes starting from p and m nodes starting from q
    # q may come from another linked list L
    def merge(self,p,n,L,q,m):
        pass
    
    # merge sort
    # sort n values starting from node
    def merge_sort(self,node=None,n=None):
        pass
    
    #sort
    def sort(self,func='selection',node=None,n=None):
        if func=='selection':
            self.selection_sort(node,n)
        elif func=='insertion':
            self.insertion_sort(node,n)
        elif func=='merge':
            self.merge_sort(node,n)
    
    #deduplicate
    # remove number of elements removed
    def deduplicate(self):
        pass
    
    # uniquify
    # return number of elements removed
    def uniquify(self):
        pass
    
    # reverse
    def reverse(self):
        pass
    
    # traverse
    def traverse(self,func):
        pass
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Sep 25 16:08:05 2022

@author: lidon
"""
import random
from vectorADT import*



# swap i and j
def swap(i,j):
    i=i+j
    j=i-j
    i=i-j
    return i,j

# copy function
def copy(self):
    new_arr=self.elements.copy()
    u=vector(new_arr)
    return u
# bind copy to vector
vector.copy=copy

# implement unsort()
# unsort elements in [low, high)
# use Knuth shuffling
def unsort(self,low=None,high=None):
    if low==None and high==None:
        low=0
        high=self.length
    for k in range(low,high):
        index=random.randint(k,high-1)
        self[k],self[index]=swap(self[k],self[index])
# bind unsort to vector
vector.unsort=unsort

#implement size()
def size(self):
    return self.length
#bind size() to vector
vector.size=size

#implement find()
# find e from [low,high)
# find the smallest index that equals to e
# if fail to find e, return -1
def find(self,e,low=None,high=None):
    if low==None and high==None:
        low=0
        high=self.length
    for k in range(low,high):
        if self[k]==e:
            return k
    return -1
# bind find to vector
vector.find=find

# insert e into position r
# return r after executing
def insert(self,r,e):
    self.elements.append(0)
    i=self.length
    # don't use add, since traversing backward have time complexity O(n-r)
    while i!=r:
        self.elements[i]=self.elements[i-1]
        i=i-1
    self.elements[i]=e
    self.length=self.length+1
    return r
#bind insert to vector
vector.insert=insert

# implement remove_index
# return number of elements deleted
# time complexity O(lengh-high), independent of high-low
def remove_index(self,low,high):
    if low==high:
        return 0
    while high<self.length:
        self.elements[low]=self.elements[high]
        low+=1
        high+=1
    self.length=low
    self.elements=self.elements[0:self.length]
    return high-low
#bind remove_index to vector
vector.remove_index=remove_index
    
# implement remove
# return element deleted
def remove(self,r):
    elem=self[r]
    self.remove_index(r,r+1)
    return elem
# bind to vector
vector.remove=remove

# implement deduplicate
# return old size - new size
def deduplicate(self):
    old_size=self.length
    # in case only 1 element
    if self.length==1:
        return 0
    r=1
    while r<self.length:
        if self.find(self.elements[r],0,r)<0:
            r=r+1
        else:
            self.remove(r)
    new_size=self.length
    return old_size-new_size
# bind to vector
vector.deduplicate=deduplicate

# implement traverse
# notice can't change elements in v
def traverse(self,func):
    for i in range(0,self.size()):
        func(self.elements[i])
# bind to vector
vector.traverse=traverse
    
#implement disordered
def disordered(self):
    n=0
    if self.length==1:
        return n
    for i in range(0,self.length-1):
        if self.elements[i]>self.elements[i+1]:
            n+=1
    return n
# bind to vector
vector.disordered=disordered

# implement uniquify
# assume v sorted
# Time complexity: O(n)
def uniquify(self):
    if self.length==1:
        return 0
    i=0
    j=1
    while j<self.length:
        if self.elements[i]!=self.elements[j]:
            self.elements[i+1]=self.elements[j]
            i=i+1
        j=j+1
    self.elements=self.elements[0:i+1]
    self.length=len(self.elements)
    return 1
# bind to vector
vector.uniquify=uniquify

#implement bin_search
# use binary search
# only applicable for ordered vector
# if search fails, return -1
def bin_search(self,e,low=None,high=None):
    if low==None and high==None:
        low=0
        high=self.length
    while low<high:
        mid=int((low+high)/2)
        if e<self.elements[mid]:
            high=mid
        elif e> self.elements[mid]:
            low=mid
        else:
            return mid
    # search failed
    return -1
# bind to vector
vector.bin_search=bin_search

# implement bubble sort
def bubble_sort(self,low=None,high=None):
    if low==None and high==None:
        low=0
        high=self.length
    for i in range(low,high):
        for j in range(low,high-i-1):
            if self.elements[j]>self.elements[j+1]:
                self.elements[j],self.elements[j+1]=swap(self.elements[j],self.elements[j+1])
# bind to vector
vector.bubble_sort=bubble_sort

#implement merge
# assume v[low:mid] and v[mid:high] are sorted, now merge them together
def merge(self,low,mid,high):
    # copy the 1st part, 2nd part doesn't need to copy in the algorithm
    # save space complexity
    temp_copy=self.elements[low:mid].copy()
    
    i=0
    j=0
    k=0
    while (j<mid-low or k<high-mid):
        if j<mid-low and (k==high-mid or temp_copy[j]<=self.elements[mid+k]):
            self.elements[low+i]=temp_copy[j]
            i+=1
            j+=1
        if k<high-mid and (j==mid-low or self.elements[mid+k]<temp_copy[j]):
            self.elements[low+i]=self.elements[mid+k]
            i+=1
            k+=1
# bind to vector
vector.merge=merge

# implement merge sort
def merge_sort(self,low=None,high=None):
    if low==None and high==None:
        low=0
        high=v.length
    if high-low<2:
        return
    mid=int((low+high)/2)
    self.merge_sort(low,mid)
    self.merge_sort(mid,high)
    self.merge(low,mid,high)
# bind to vector
vector.merge_sort=merge_sort   
    
# return maximum
def maximum(self,low=None,high=None):
    if low==None and high==None:
        low=0
        high=self.length
    output=self.elements[0]
    for i in range(0,self.length):
        if self.elements[i]>output:
            output=self.elements[i]
    return output
#bind to vector
vector.maximum=maximum
    
